#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/serial.h>
#include <linux/serial_core.h>
#include <linux/tty_flip.h>
#include "my_uart.h"

static unsigned int my_uart_tx_empty(struct uart_port *port);
static unsigned int my_uart_get_mctrl(struct uart_port *port);
static void my_uart_set_mctrl(struct uart_port *port, unsigned int m);
static void my_uart_stop_tx(struct uart_port *port, unsigned int c);
static void my_uart_start_tx(struct uart_port *port, unsigned int start);
static void my_uart_stop_rx(struct uart_port *port);
static void my_uart_enable_ms(struct uart_port *port);
static void my_uart_break_ctl(struct uart_port *uport, int break_state);
static int my_uart_startup(struct uart_port *port);
static void my_uart_shutdown(struct uart_port *port);
static void my_uart_set_termios(struct uart_port *port, struct termios *termios, struct termios *old);
static const char *my_uart_type(struct uart_port *port);
static void my_uart_release_port(struct uart_port *port);
static int my_uart_request_port(struct uart_port *port);
static void my_uart_config_port(struct uart_port *port, int flags);
static int my_uart_verify_port(struct uart_port *port, struct serial_struct *ser);
static irqreturn_t my_uart_irq_handler(int irq, void *dev_id, struct pt_regs *regs);

static struct uart_ops my_uart_ops= {
        .tx_empty       = my_uart_tx_empty,
        .get_mctrl      = my_uart_get_mctrl,
        .set_mctrl      = my_uart_set_mctrl,
        .stop_tx        = my_uart_stop_tx,
        .start_tx       = my_uart_start_tx,
        .stop_rx        = my_uart_stop_rx,
        .enable_ms      = my_uart_enable_ms,
        .break_ctl      = my_uart_break_ctl,
        .startup        = my_uart_startup,
        .shutdown       = my_uart_shutdown,
        .set_termios    = my_uart_set_termios,
        .type           = my_uart_type,
        .release_port   = my_uart_release_port,
        .request_port   = my_uart_request_port,
        .config_port    = my_uart_config_port,
        .verify_port    = my_uart_verify_port,
};

static struct uart_driver my_uart_driver = {
	.owner = THIS_MODULE,
	.driver_name = "serial",
	.dev_name = "ttyS0",
	.major = TTY_MAJOR,
	.minor = 1,
	.nr = 1,
};

static struct uart_port my_uart_port = {
	.membase = MY_UART_MEMBASE,
	.iotype = SERIAL_IO_MEM,
	.irq = MY_UART_IRQ,
	.fifosize = 32,
	.line = 0,
	.ops = &my_uart_ops
};

int __init my_uart_init(void)
{
	/* uart_register_driver binds the low level driver with the serial CORE which turn
	 * registers with the TTY layer using the tty_register_driver()  function. Also the
	 * uart_state structures are creates.(the  number of these structures is equivalent
	 * to number of hardward ports) and  pointer to this array is stored in my_uart_driver.
	*/
	uart_register_driver (&my_uart_driver);

	/* As indicated in the above figure this function connects the uart_state to the
	 * uart_port. Also this function lets the TTY layer know that a device has been added
	 * using the function tty_register_device(). 
	 */
	uart_add_one_port (&my_uart_driver, &my_uart_port);
	return 0;
}


static int my_uart_startup(struct uart_port *port)
{
	return(request_irq(MY_UART_IRQ, my_uart_irq_handler, 0,"SAMPLE UART", &my_uart_port));
}

static void my_uart_shutdown(struct uart_port *port)
{
	free_irq(MY_UART_IRQ, port);
}


static void my_uart_char_tx(struct uart_port *port)
{ 
	struct circ_buf *xmit = &port->info->xmit;
	
	/* If a XON/XOFF character needs to be transmitted out, the x_char field of the port is set by the serial core */
	if(port->x_char)
	{
		SEND_CHAR(port->x_char);
		port->x_char = 0; /* Reset the field */
		return;	
	}

	if(uart_tx_stopped(port) || uart_circ_empty(xmit))
	{
		my_uart_stop_tx(port, 0);
		return;
	}

	SEND_CHAR(xmit->buf[xmit->tail]);
	/* UART_XMIT_SIZE is defined in include/linux/serial_core.h */
	xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE + 1 );

	/* Now check if there are more characters that need to be sent and we have enough space
	 * in the transmission buffer which is defined by the macro WAKEUP_CHARS set to 256 in
	 * the file include/linux/serial_core.h. The function uart_write_wakeup provided by the
	 * serial core ultimately ends up calling the TTY wakeup handler function which in turn
	 * informs the line discipline that the low level driver is ready to receive more data. 
	 */
	if(uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(port);

	if(uart_circ_empty(xmit))
		my_uart_stop_tx(port, 0);
}

static void my_uart_stop_tx(struct uart_port *port, unsigned int c)
{
	STOP_TX();
}

static void my_uart_start_tx(struct uart_port *port, unsigned int start)
{
	START_TX();
	my_uart_char_tx(port);
}

/* Return 0 if not empty */
static unsigned int my_uart_tx_empty(struct uart_port *port)
{
	return (TX_EMPTY()? TIOCSER_TEMT : 0);
}

void my_uart_char_rx(struct uart_port *port)
{
	struct tty_struct *tty = port->info->tty;
	struct uart_icount *icount = &port->icount;
	unsigned int i=0;
	
	while(	CHAR_READY())
	{
		unsigned char c;
		unsigned char flag = TTY_NORMAL;
		unsigned char st = READ_RX_STATUS();

		if(tty->flip.count >= TTY_FLIPBUF_SIZE)
			break;

		c = READ_CHAR();

		icount->rx++;

		if(st & (PARITY_ERROR | FRAME_ERROR | OVERRUN_ERROR) )
		{
			if(st & PARITY_ERROR)
				icount->parity ++;
			if(st & FRAME_ERROR)
				icount->frame ++;
			if(st & OVERRUN_ERROR)
				icount->overrun ++;

			/* If we have been asked to ignore errors then do the following */
			if(st & port->ignore_status_mask)
			{
				if(++i > IGNORE_ERROR_NUM)
					break;
				goto ignore;
			}
			/* Report the errors that we have not been asked to ignore */
			st &= port->read_status_mask; 
			if(st & PARITY_ERROR) flag = TTY_PARITY;
			if(st & FRAME_ERROR) flag = TTY_FRAME;
			/* Overrun is a special case; it does not affect the character read */
			if(st & OVERRUN_ERROR) 
			{
				tty_insert_flip_char(tty, c, flag);
				c = 0;
				flag = TTY_OVERRUN;
			}
		}
		tty_insert_flip_char(tty,c,flag);
ignore:
	}

	tty_flip_buffer_push(tty); 
}

static void my_uart_stop_rx(struct uart_port *port)
{
	STOP_RX();
}


static irqreturn_t my_uart_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned int st= READ_INTERRUPT_STATUS();

	if(st & TX_INT_MASK) my_uart_char_tx(&my_uart_port);
	if(st & RX_INT_MASK) my_uart_char_rx(&my_uart_port);

	return IRQ_HANDLED;
}

//Since there is no modem support the following functions are implemented as dummy functions

static void my_uart_set_mctrl(struct uart_port *port, unsigned int m)
{
	return;
}

static unsigned int my_uart_get_mctrl(struct uart_port *port)
{
	return TIOCM_CAR | TIOCM_DSR | TIOCM_CTS;
}


static void my_uart_set_termios(struct uart_port *port, struct termios *termios, struct termios *old)
{
	unsigned int c_cflag = termios->c_cflag;
	unsigned int baud=9600, stop_bits=1, parity=0, data_bits=8;
	unsigned long flags;
	
	/* Calculate the number of data bits */
	switch (c_cflag & CSIZE) {
		case CS5: data_bits = 5; break;
		case CS6: data_bits = 6; break;
		case CS7: data_bits = 7; break;
		case CS8: data_bits = 8; break;
		default: data_bits = 8;
	}

	if(c_cflag & CSTOPB) stop_bits = 2;

	if(c_cflag & PARENB) parity = 1;
	if(c_cflag & PARODD) parity = 2;

	/* We support only 2 speeds of 9600 and 19200. Translate the termios settings into any one of these */
	baud = uart_get_baud_rate(port, termios, old, 9600, 19200);

	spin_lock_irqsave(&port->lock, flags);
	SET_SPEED(baud);
	SET_STOP_BITS(stop_bits);
	SET_PARITY(parity);
	SET_BITS(data_bits);
	port->read_status_mask = OVERRUN_ERROR;
	if(termios->c_iflag & INPCK)
		port->read_status_mask |= PARITY_ERROR | FRAME_ERROR;
	port->ignore_status_mask = 0;
	if(termios->c_iflag & IGNPAR)
		port->ignore_status_mask |= PARITY_ERROR | FRAME_ERROR;
	spin_unlock_irqrestore(&port->lock, flags);
}

static const char *my_uart_type(struct uart_port *port)
{
        return "SAMPLE";
}

static void my_uart_enable_ms(struct uart_port *port)
{
	return;
}

static void my_uart_break_ctl(struct uart_port *uport, int break_state)
{
	return;
}

static void my_uart_release_port(struct uart_port *port)
{
	return; 
}

static int my_uart_request_port(struct uart_port *port)
{
	return; 
}

static void my_uart_config_port(struct uart_port *port, int flags)
{
	return;
}

static int my_uart_verify_port(struct uart_port *port, struct serial_struct *ser)
{
        return 0;
}


MODULE_DESCRIPTION("SAMPLE serial driver");
MODULE_LICENSE("GPL");

